#ifndef GFX_ENTITY_OBJECT_HPP
#define GFX_ENTITY_OBJECT_HPP

#include "gfx/entity.hpp"

#include <vector>

namespace gfx
{
	class EntityObject;
	class Mesh;
	class Shader;

	/** Convenience typedef. */
	typedef boost::shared_ptr<EntityObject> EntityObjectSptr;

	/** \brief Normal object.
	 *
	 * Normal objects are essentially recursive trees of themselves. An object
	 * contains any number of meshes and subsequently any number of recursive
	 * objects. The recursive object references are transformed downward from
	 * the original orientation of the roof entity.
	 */
	class EntityObject : public Entity
	{
		protected:
			/** Meshes.*/
			std::vector<const Mesh*> m_meshes;

			/** Recursive objects. */
			std::vector<EntityObjectSptr> m_objects;

			/** Do we want to draw this or not. */
			bool draw_disabled;

		public:
			/** \brief Construtor. */
			EntityObject();

			/** \brief Constructor.
			 *
			 * @param ppos Position.
			 * @param prot Rotation.
			 * @param pmesh mesh to use.
			 */
			EntityObject(const math::vec3d &ppos, const math::vec3d &prot);

			/** \brief Constructor.
			 *
			 * @param ppos Position.
			 * @param prot Rotation.
			 * @param pmesh mesh to use.
			 */
			EntityObject(const math::vec3d &ppos, const math::vec3d &prot,
					const Mesh *pmesh);

			/** \brief Constructor.
			 *
			 * @param pmesh mesh to use.
			 */
			EntityObject(const Mesh *pmesh);

			/** \brief Destructor.
			 *
			 * Although the meshes are pointers from other sources and are not
			 * destroyed when the object is destroyed, all recursive objects are.
			 */
			virtual ~EntityObject() { }

		public:
			/** \brief Draw this object.
			 *
			 * Also draws all other objects recursively.
			 *
			 * @param psh Shader to use.
			 * @param pmat Modelview matrix to use.
			 */
			virtual void draw(const Shader &psh, const math::mat4f &pmat);

			/** \brief Query visibility of the object.
			 *
			 * @return Status. true if drawing is disabled, false if not.
			 */
			bool isDrawDisabled();

		protected:
			/** \brief Prepares the orientation for this object
			 *
			 * The default implementation just multiplies the object matrix into the
			 * matrix stack.
			 *
			 * @param pmat Matrix currently on stack.
			 */
			virtual math::mat4f orient(const math::mat4f &pmat);

		public:
			/** \brief Add a mesh to this.
			 *
			 * @param pmesh Mesh to add.
			 */
			inline void addMesh(const Mesh *pmesh)
			{
				m_meshes.push_back(pmesh);
			}

			/** \brief Adjust visibility of the object - show.
			 */
			inline void enableDraw()
			{
				draw_disabled = false;
			}

			/** \brief Adjust visibility of the object - hide.
			 */
			inline void disableDraw()
			{
				draw_disabled = true;
			}
	};
}

#endif
