#include "gfx/mesh.hpp"

#include "data/generic.hpp"
#include "thr/dispatch.hpp"
#include "ui/generic.hpp"

#include <boost/filesystem.hpp>
#include <boost/property_tree/xml_parser.hpp>

using namespace gfx;
namespace fs = boost::filesystem;
namespace pt = boost::property_tree;

Mesh::Mesh(const std::string &pfname)
{
	this->load(pfname);
}

void Mesh::addTexture(const std::string &id, void *tex)
{
	if(id.compare(0, 7, "texture"))
	{
		std::stringstream sstr;
		sstr << "unknown texture id: " << id;
		BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
	}
	_textures.push_back(Texture2DSptr(static_cast<Texture2D*>(tex)));
}

void Mesh::calcNormals()
{
	unsigned vsize = static_cast<unsigned>(_vertex.size());
	_normal.resize(0);
	_normal.resize(vsize, math::vec3f(0.0f, 0.0f, 0.0f));
	std::vector<unsigned> refcnts;;
	refcnts.resize(vsize, 0);

	TriVec faces;
	_lod.collect(faces);

	BOOST_FOREACH(const Triangle &tri, faces)
	{
		math::vec3f nor = this->calcNormal(tri);
		unsigned ia = tri.a(),
						 ib = tri.b(),
						 ic = tri.c();
		_normal[ia] += nor;
		_normal[ib] += nor;
		_normal[ic] += nor;
		++(refcnts[ia]);
		++(refcnts[ib]);
		++(refcnts[ic]);
	}

	for(unsigned ii = 0; (ii < vsize); ++ii)
	{
		math::vec3f &nn = _normal[ii];

		unsigned rcnt = refcnts[ii];
		if(rcnt > 0)
		{
			nn = math::normalize(nn / static_cast<float>(rcnt));
		}
		else
		{
			// Fallback normal points up.
			nn = math::vec3f(0.0f, 1.0f, 0.0f);
		}
	}

	data::stl_vector_trim(_color);
	data::stl_vector_trim(_texcoord);
	data::stl_vector_trim(_vertex);
}

math::rect3f Mesh::getBoundary() const
{
	if(_vertex.size() <= 0)
	{
		std::stringstream sstr;
		sstr << "boundary requested for mesh that has no vertices";
		BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
	}
	math::rect3f ret(_vertex[0]);
	for(unsigned ii = 1; (ii < _vertex.size()); ++ii)
	{
		ret.expand(_vertex[ii]);
	}
	return ret;
}

void Mesh::load(const std::string &pfname, float pscale)
{
	math::vec3f obj_scale(pscale, pscale, pscale);
	bool enable_center = false,
			 enable_into = false,
			 enable_scale = (pscale != 1.0f);

	this->unreserve();

	ui::log(std::string("loading mesh '") + pfname + std::string("'"));

	data::shristr pfile(data::open_read(pfname));
	pt::ptree xtree;
	pt::read_xml(*pfile, xtree);

	pt::ptree::value_type root = xtree.front();
	if(root.first.compare("mesh"))
	{
		std::stringstream err;
		err << "unknown root type: " << root.first;
		BOOST_THROW_EXCEPTION(std::invalid_argument(err.str()));
	}

	BOOST_FOREACH(pt::ptree::value_type &vv, root.second)
	{
		const std::string &type = vv.first;
		const pt::ptree &subtree = vv.second;

		if(!type.compare("scale"))
		{
			enable_scale = true;
			enable_center = subtree.get<bool>("center", 0);
			enable_into = subtree.get<bool>("into", 1);
			obj_scale.x() = subtree.get<float>("x", 1.0f);
			obj_scale.y() = subtree.get<float>("y", 1.0f);
			obj_scale.z() = subtree.get<float>("z", 1.0f);
		}
		else if(!type.compare("vertex"))
		{
			this->readVertex(&subtree);
		}
		else if(!type.compare("face"))
		{
			// Must flip faces for correct orientation.
			_lod.addFace(Triangle(subtree.get<unsigned>("a"),
						subtree.get<unsigned>("c"),
						subtree.get<unsigned>("b")));
		}
		else if((type.compare(0, 7, "texture") == 0) ||
				(type.compare(0, 9, "normalmap") == 0))
		{
			fs::path subpath(fs::path(pfname).parent_path().parent_path() /
					subtree.get<std::string>(""));
			thr::wait_privileged(boost::bind(&Mesh::taskTextureFile, this,
						type.c_str(), boost::ref(subpath.file_string())));
		}
		else if(type.compare(0, 6, "volume") == 0)
		{
			std::stringstream sstr;
			sstr << "can not load volume: " << type;
			BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
		}
	}

	if(enable_scale)
	{
		if(enable_into || enable_center)
		{
			math::rect3f area(_vertex[0]);
			for(unsigned ii = 1; (ii < _vertex.size()); ++ii)
			{
				area.expand(_vertex[ii]);
			}
			//std::cout << "model area: " << area << std::endl;

			// Center before scaling for easiness.
			if(enable_center)
			{
				math::vec3f center = area.center();
				BOOST_FOREACH(math::vec3f &vv, _vertex)
				{
					vv -= center;
				}
				area = math::rect3f(area.getMin() - center, area.getMax() - center);
				//std::cout << "centered: " << area << std::endl;
			}

			// Scale into after potential centering.
			if(enable_into)
			{
				math::vec3f size = area.getMax() - area.getMin();
				obj_scale.x() = obj_scale.x() / size.x();
				obj_scale.y() = obj_scale.y() / size.y();
				obj_scale.z() = obj_scale.z() / size.z();
			}
		}
		this->scale(math::min(obj_scale.x(), obj_scale.y(), obj_scale.z()));
	}

	this->compile();
}

void Mesh::readVertex(const void *subtree)
{
	const pt::ptree *st = reinterpret_cast<const pt::ptree*>(subtree);

	try {
		_bone_ref.push_back(math::vec4u(st->get<unsigned>("bone_ref.a"),
					st->get<unsigned>("bone_ref.b"),
					st->get<unsigned>("bone_ref.c"),
					st->get<unsigned>("bone_ref.d")));
	} catch(...) { }
	try {
		_bone_weight.push_back(math::vec4f(st->get<float>("bone_weight.a"),
					st->get<float>("bone_weight.b"),
					st->get<float>("bone_weight.c"),
					st->get<float>("bone_weight.d")));
	} catch(...) { }
	try {
		_color.push_back(Color(st->get<float>("color.r"),
					st->get<float>("color.g"),
					st->get<float>("color.b"),
					st->get<float>("color.a")));
	} catch(...) { }
	try {
		_normal.push_back(math::vec3f(st->get<float>("normal.x"),
					st->get<float>("normal.y"),
					st->get<float>("normal.z")));
	} catch(...) { }
	try {
		_texcoord.push_back(math::vec2f(st->get<float>("texcoord.s"),
					st->get<float>("texcoord.t")));
	} catch(...) { }
	// Vertex MUST exist, others are voluntary.
	math::vec3f ver(st->get<float>("x"),
				st->get<float>("y"),
				st->get<float>("z"));
	// Must 'rotate' to correct orientation.
	_vertex.push_back(math::vec3f(ver.y(), ver.z(), -ver.x()));
}

void Mesh::scale(const math::vec3f &svec)
{
	BOOST_FOREACH(math::vec3f &vv, _vertex)
	{
		vv *= svec;
	}
}

void Mesh::taskElem()
{
	_elem.update(_lod);
}

void Mesh::taskTextureFile(const char *type, const std::string &pfname)
{
	this->addTexture(std::string(type), new Texture2D(pfname));
}

void Mesh::taskTextureGray8(const char *type, const ImageGray8 *img)
{
	this->addTexture(std::string(type), new Texture2D(*img, true));
}

void Mesh::taskTextureGray16(const char *type, const ImageGray16 *img)
{
	this->addTexture(std::string(type), new Texture2D(*img, true));
}

void Mesh::taskTextureRGB(const char *type, const ImageRGB *img)
{
	this->addTexture(std::string(type), new Texture2D(*img, true));
}

void Mesh::unreserve()
{
	_bone_ref.clear();
	_bone_weight.clear();
	_color.clear();
	_normal.clear();
	_texcoord.clear();
	_vertex.clear();
	_lod.unreserve();
	_textures.clear();
}

