#ifndef GFX_MESH_HPP
#define GFX_MESH_HPP

#include "gfx/lod.hpp"
#include "gfx/texture_2d.hpp"
#include "math/mat.hpp"

#include <vector>

namespace gfx
{
	class Shader;

	/** \brief A basic mesh class.
	 *
	 * Each mesh is either static or animated. The base class contains data
	 * relevant to both of the variants.
	 *
	 * Note that if some of the data is not present, the mesh is just plain empty.
	 */
	class Mesh
	{
		protected:
			/** Attribute data array. */
			std::vector<math::vec4u> _bone_ref;

			/** Attribute data array. */
			std::vector<math::vec4f> _bone_weight;

			/** Attribute data array. */
			std::vector<Color> _color;

			/** Attribute data array. */
			std::vector<math::vec3f> _normal;

			/** Attribute data array. */
			std::vector<math::vec2f> _texcoord;

			/** Attribute data array. */
			std::vector<math::vec3f> _vertex;

			/** Face data. */
			Lod _lod;

			/** Element buffer. */
			BufferElem _elem;

			/** Textures in this mesh. */
			std::vector<Texture2DSptr> _textures;

		public:
			/** \brief Empty constructor. */
			Mesh() { }

			/** \brief Load constructor.
			 *
			 * @param pfname Filename to load from.
			 */
			Mesh(const std::string &pfname);

			/** \brief Destructor. */
			virtual ~Mesh() { }

		public:
			/** \brief Get the boundary of this mesh.
			 *
			 * @return Boundary calculated from this.
			 */
			math::rect3f getBoundary() const;

			/** \brief Load this from a file.
			 *
			 * Unknown elements will be passed into a specific interpretation
			 * function that is implemented in inheriting classes.
			 *
			 * @param pfname Filename to load.
			 * @param pscale Scale factor.
			 */
			void load(const std::string &pfname, float pscale = 1.0f);

			/** \brief Scale this object.
			 *
			 * Multiplies all vertices (but nothing else) with the given values.
			 *
			 * @param svec Scale vector.
			 */
			void scale(const math::vec3f &svec);

			/** \brief Element update task.
			 */
			void taskElem();

			/** \brief Texture task (filename).
			 *
			 * @param type Type to load as.
			 * @param pfname Filename to load from.
			 */
			void taskTextureFile(const char *type, const std::string &pfname);

			/** \brief Texture task (RGB image).
			 *
			 * @param type Type to load as.
			 * @param img Image to use.
			 */
			void taskTextureGray8(const char *type, const ImageGray8 *img);

			/** \brief Texture task (RGB image).
			 *
			 * @param type Type to load as.
			 * @param img Image to use.
			 */
			void taskTextureGray16(const char *type, const ImageGray16 *img);

			/** \brief Texture task (RGB image).
			 *
			 * @param type Type to load as.
			 * @param img Image to use.
			 */
			void taskTextureRGB(const char *type, const ImageRGB *img);

		protected:
			/** \brief Calculate normals.
			 *
			 * Calculates normals from vertex and face data.
			 */
			void calcNormals();

		public:
			/** \brief Add one texture into this.
			 *
			 * Virtual function to handle adding a new texture.
			 *
			 * The id contains the type or detailed name of the texture that was either
			 * loaded from an XML file or specified explicitly. Its processing may be
			 * omitted at will. The default loader accepts strings starting with
			 * "texture". Further implementations may also use such strings as
			 * "normalmap" or "volume".
			 *
			 * The parameter is passed as void pointer due to type problems for the
			 * Texture inheritance. The default implementation simply interprets the
			 * pointer as Texture2D and pushes it into the texture array of this.
			 *
			 * @param id Identifier.
			 * @param tex Texture.
			 */
			virtual void addTexture(const std::string &id, void *tex);

			/** \brief Draw this.
			 *
			 * It is expected that the camera matrix be loaded into the OpenGL
			 * stack prior to calling this method.
			 *
			 * @param psh Shader to use.
			 * @param pmat Modelview matrix to use.
			 */
			virtual void draw(const Shader &psh, const math::mat4f &op) const = 0;

		protected:
			/** \brief Compile this to be ready for drawing.
			 */
			virtual void compile() = 0;

			/** \brief Read one vertex from a property tree.
			 *
			 * Inheriting classes should call this base method.
			 *
			 * @param subtree Tree to read from.
			 */
			virtual void readVertex(const void *subtree);

			/** \brief Free all resources required by the font.
			 * 
			 * Should be done before deinitializing OpenGL.
			 *
			 * Inheriting classes should call this base method.
			 */
			virtual void unreserve();

		public:
			/** \brief Get the face normal for a given face.
			 *
			 * @param tri Triangle to use.
			 * @return Resulting normal.
			 */
			inline const math::vec3f calcNormal(const Triangle &tri)
			{
				const math::vec3f &va = _vertex[tri.a()];

				return math::cross(_vertex[tri.b()] - va, _vertex[tri.c()] - va);
			}

			/** \brief Scale this object.
			 *
			 * Multiplies all vertices (but nothing else) with the given value.
			 *
			 * @param op Scale parameter.
			 */
			inline void scale(float op)
			{
				this->scale(math::vec3f(op, op, op));
			}
	};

	/** Conevnience typedef. */
	typedef boost::shared_ptr<Mesh> MeshSptr;
}

#endif
