#include "gfx/mesh_planet.hpp"

#include "data/generic.hpp"
#include "gfx/height_map_ball.hpp"
#include "gfx/shader.hpp"
#include "ui/generic.hpp"
#include "thr/dispatch.hpp"

using namespace gfx;

MeshPlanet::MeshPlanet(unsigned subdivision,
		unsigned subdivision_coalesce, const HeightMapBall *hmap,
		unsigned texture_detail)
{
	this->compile(subdivision, subdivision_coalesce, hmap, texture_detail);
}

void MeshPlanet::addTexture(const std::string &id, void *tex)
{
	if(id.compare("texture") == 0)
	{
		_textures.push_back(Texture2DSptr(static_cast<Texture2D*>(tex)));
		return;
	}
	if(id.compare("normalmap") == 0)
	{
		_normalmaps.push_back(Texture2DSptr(static_cast<Texture2D*>(tex)));
		return;
	}
	if(id.compare("volume") == 0)
	{
		_volumes.push_back(Texture3DSptr(static_cast<Texture3D*>(tex)));
		return;
	}

	std::stringstream sstr;
	sstr << "unknown texture type: " << id;
	BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
}

void MeshPlanet::compile()
{
	this->compile(0);
}

void MeshPlanet::compile(unsigned subdivision,
		unsigned subdivision_coalesce, const HeightMapBall *hmap,
		unsigned texture_detail, unsigned volume_detail,
		const std::string &fname_header, bool psave)
{
	this->subdivide(subdivision, true);

	// Height mapping phase. The image loads will throw an exception, and either
	// all images are specified or none are.
	if(hmap)
	{
		for(unsigned kk = 0; (kk < 20); kk += 2)
		{
			std::stringstream sstream_fname_tex, sstream_fname_nor;
			ImageGray16Sptr ctex;
			ImageRGBSptr ntex;

			sstream_fname_tex << fname_header << "_tex_" << texture_detail <<
				"_" << (kk / 2) << ".png";
			sstream_fname_nor << fname_header << "_nor_" << texture_detail <<
				"_" << (kk / 2) << ".png";
			try
			{
				std::string fname_tex = data::open_search(sstream_fname_tex.str()),
					fname_nor = data::open_search(sstream_fname_nor.str());
				ctex = ImageGray16Sptr(new ImageGray16(fname_tex));
				ntex = ImageRGBSptr(new ImageRGB(fname_nor));
			}
			catch(...)
			{
				ctex = ImageGray16Sptr(new ImageGray16(texture_detail, texture_detail));
				ntex = ImageRGBSptr(new ImageRGB(texture_detail, texture_detail));
				{
					std::stringstream sstr;
					sstr << "creating planet facets " << kk << " and " << (kk + 1);
					ui::log(sstr.str());
				}

				Lod &lod1 = *(_lod.getRecursive()[kk + 0]),
						&lod2 = *(_lod.getRecursive()[kk + 1]);
				Triangle &tt1 = lod1.getFaces().front(),
								 &tt2 = lod2.getFaces().front();
				if((lod1.getFaces().size() > 1) || (lod2.getFaces().size() > 1))
				{
					BOOST_THROW_EXCEPTION(std::runtime_error("non-single LOD unit"));
				}
				if((tt1.b() != tt2.b()) || (tt1.c() != tt2.a()))
				{
					BOOST_THROW_EXCEPTION(std::runtime_error("malformed triangle"));
				}
				math::vec3f &vv1 = _vertex[tt1.a()],
					&vv2 = _vertex[tt1.b()],
					&vv3 = _vertex[tt1.c()],
					&vv4 = _vertex[tt2.c()],
					vx1(vv2 - vv1),
					vy1(vv3 - vv1),
					vx2(vv3 - vv4),
					vy2(vv2 - vv4);

				float fwidth = static_cast<float>(ctex->getWidth() - 1),
							fheight = static_cast<float>(ctex->getHeight() - 1);
				for(unsigned ii = 0; (ii < ctex->getWidth()); ++ii)
				{
					float fi = static_cast<float>(ii) / fwidth;

					for(unsigned jj = 0; (jj < ctex->getHeight()); ++jj)
					{
						float fj = static_cast<float>(jj) / fheight;

						math::vec3f vv;
						if(fi + fj < 1.0f)
						{
							vv = (fi * vx1 + fj * vy1) + vv1;
						}
						else
						{
							vv = ((1.0f - fi) * vx2 + (1.0f - fj) * vy2) + vv4;
						}

						float diff = hmap->calcHeight(vv);
						ctex->setPixel(ii, jj, hmap->calcColor(vv, diff));

						math::vec3f nn = hmap->calcNormal(vv, 1.0f / (static_cast<float>(texture_detail) * 0.125f), diff);
						ntex->setPixel(ii, jj, nn * 0.5f + 0.5f);
					}
				}
				if(psave)
				{
					std::stringstream sstr;
					sstr << "saving '" << sstream_fname_tex.str() << "'";
					ui::log(sstr.str());
					ctex->write(sstream_fname_tex.str());
					
					sstr.str("");
					sstr << "saving '" << sstream_fname_nor.str() << "'";
					ui::log(sstr.str());
					ntex->write(sstream_fname_nor.str());
				}
			}
			thr::dispatch_privileged(boost::bind(&Mesh::taskTextureGray16, this,
						"texture", ctex.get()));
			thr::wait_privileged(boost::bind(&Mesh::taskTextureRGB, this,
						"normalmap", ntex.get()));
		}

		// Vertex height set phase after creating the correct polygons.
		for(unsigned ii = 0; (ii < _vertex.size()); ++ii)
		{
			_vertex[ii] = hmap->normalizeHeight(_vertex[ii]);
		}
	}

	this->coalesce(subdivision_coalesce, subdivision);
	this->createVolumes(fname_header, volume_detail, psave);

	_color.clear();
	_normal.clear();
	data::stl_vector_trim(_color);
	data::stl_vector_trim(_normal);
	data::stl_vector_trim(_texcoord);
	data::stl_vector_trim(_vertex);

	thr::dispatch_privileged(boost::bind(&MeshPlanet::taskBuffer, this));
	thr::wait_privileged(boost::bind(&Mesh::taskElem, this));
	_lod.compile(_vertex);
}

void MeshPlanet::taskBuffer()
{
	_buf.update(_texcoord, _vertex);
}

void MeshPlanet::taskVolumeGray8(const VolumeGray8 &volume)
{
	this->addTexture("volume", new Texture3D(volume));
}

void MeshPlanet::taskVolumeRGB(const VolumeRGB &volume)
{
	this->addTexture("volume", new Texture3D(volume));
}

void MeshPlanet::taskVolumeRGBA(const VolumeRGBA &volume)
{
	this->addTexture("volume", new Texture3D(volume));
}

void MeshPlanet::unreserve()
{
	_buf.unreserve();
	_normalmaps.clear();
	_volumes.clear();
	this->Mesh::unreserve();
}

