#include "ob_globals.hpp"

#include "gfx/mesh_static.hpp"
#include "thr/dispatch.hpp"
#include "ob_atmosphere.hpp"
#include "ob_constants.hpp"
#include "ob_height_map_planet.hpp"
#include "ob_menu_state.hpp"
#include "ob_planet.hpp"
#include "ob_visualization_flak.hpp"
#include "ob_visualization_nuke.hpp"
#include "ob_visualization_orbit.hpp"
#include "ob_visualization_railgun.hpp"

#include <sstream>

using namespace ob;

bool Globals::generate_enabled = false;

Game *ob::game = NULL;
Globals *ob::glob = NULL;

/** \brirf Log to console.
 *
 * @param op String to log.
 */
static void log_console(const std::string &op)
{
	glob->getConsole().addRow(ui::wstr_utf8(op));
}

Globals::Globals(const gfx::SurfaceScreen &pscreen, const char *pdetail) :
	_detail_level(pdetail),
	_font("fnt/default.xml"),
	_console(_font, OB_CONSOLE_FONT_SIZE, pscreen),
	_cursor_blank(create_cursor_blank()),
	_cursor_default(SDL_GetCursor()),
	m_menu_game(MenuState::create_menu_game()),
	m_menu_main(MenuState::create_menu_main()),
	_precalculated(false)
{
	_console.setProgress(0);
	_console.setProgressMax(100);
}

Globals::Globals(const gfx::SurfaceScreen &pscreen, const std::string &pdetail) :
	_detail_level(pdetail),
	_font("fnt/default.xml"),
	_console(_font, OB_CONSOLE_FONT_SIZE, pscreen),
	_cursor_blank(create_cursor_blank()),
	_cursor_default(SDL_GetCursor()),
	_precalculated(false)
{
	_console.setProgress(0);
	_console.setProgressMax(100);
}

Globals::~Globals()
{
	this->unreserve();
	SDL_SetCursor(_cursor_default);
	SDL_FreeCursor(_cursor_blank);
}

void Globals::precalc()
{
	if(_detail_level.compare("laptop") == 0)
	{
		this->precalc(6, 3, 256, 64);
		return;
	}
	
	if(_detail_level.compare("desktop") == 0)
	{
		this->precalc(7, 3, 1024, 128);
		return;
	}

	if(_detail_level.compare("bleeding") == 0)
	{
		this->precalc(8, 3, 2048, 256);
		return;
	}

	if(_detail_level.compare("custom") != 0)
	{
		std::stringstream sstr;
		sstr << "unknown detail level: " << _detail_level;
		BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
	}

	// TODO: implement.
	BOOST_THROW_EXCEPTION(std::runtime_error("unimplemented"));
}

void Globals::precalc(unsigned subdivide, unsigned coalesce, unsigned texsize,
		unsigned volsize)
{
	thr::wait_privileged(boost::bind(&Globals::unreserve, this));
	
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_atmosphere), "shader/3d_atmosphere.xml"));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_object), "shader/3d_static.xml"));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_overlay), "shader/3d_overlay.xml"));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_overlay_billboard), "shader/3d_overlay_billboard.xml"));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_overlay_line), "shader/3d_overlay_line.xml"));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_planet), "shader/3d_world.xml"));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_shader,
				boost::ref(m_shader_planet_schematic), "shader/3d_world_schematic.xml"));

	gfx::ImageRGBA img_flak_ammo("gfx/textures/texture_flak_ammo.png"),
		img_icon_bullet_flak("gfx/textures/icon_bullet_flak.png"),
		img_icon_bullet_railgun("gfx/textures/icon_bullet_railgun.png"),
		img_icon_city("gfx/textures/icon_city.png"),
		img_icon_missile_anti_nuke("gfx/textures/icon_missile_anti_nuke.png"),
		img_icon_missile_anti_ship("gfx/textures/icon_missile_anti_ship.png"),
		img_icon_silo("gfx/textures/icon_silo.png"),
		img_missile_anti_nuke("gfx/billboards/billboard_missile_anti_nuke.png"),
		img_missile_anti_nuke_warning("gfx/billboards/billboard_warning_yellow_frame_1.png"),
		img_missile_anti_ship("gfx/billboards/billboard_missile_anti_ship.png"),
		img_missile_anti_ship_warning("gfx/billboards/billboard_warning_red_frame_1.png"),
		img_missile_nuke("gfx/billboards/billboard_missile_nuke.png"),
		img_nuke_ammo("gfx/textures/texture_nuke_ammo.png"),
		img_railgun_ammo("gfx/textures/texture_railgun_ammo.png"),
		img_reload("gfx/textures/texture_reload.png"),
		img_silo_anti_nuke("gfx/billboards/billboard_silo_anti_nuke.png"),
		img_silo_anti_ship("gfx/billboards/billboard_silo_anti_ship.png"),
		img_silo_both("gfx/billboards/billboard_silo_both.png"),
		img_target("gfx/billboards/billboard_target.png"),
		img_trace("gfx/textures/texture_trace.png");
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_flak_ammo), boost::ref(img_flak_ammo), false));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_icon_bullet_flak), boost::ref(img_icon_bullet_flak), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_icon_bullet_railgun), boost::ref(img_icon_bullet_railgun), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_icon_city), boost::ref(img_icon_city), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_icon_missile_anti_nuke), boost::ref(img_icon_missile_anti_nuke), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_icon_missile_anti_ship), boost::ref(img_icon_missile_anti_ship), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_icon_silo), boost::ref(img_icon_silo), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_missile_anti_nuke), boost::ref(img_missile_anti_nuke), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_missile_anti_nuke_warning), boost::ref(img_missile_anti_nuke_warning), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_missile_anti_ship), boost::ref(img_missile_anti_ship), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_missile_anti_ship_warning), boost::ref(img_missile_anti_ship_warning), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_missile_nuke), boost::ref(img_missile_nuke), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_nuke_ammo), boost::ref(img_nuke_ammo), false));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_railgun_ammo), boost::ref(img_railgun_ammo), false));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_reload), boost::ref(img_reload), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_silo_anti_nuke), boost::ref(img_silo_anti_nuke), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_silo_anti_ship), boost::ref(img_silo_anti_ship), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_silo_both), boost::ref(img_silo_both), true));
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_target), boost::ref(img_target), true));
	// Trace line must loop.
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
				boost::ref(m_texture_trace), boost::ref(img_trace), false));

	gfx::ImageRGB img_menu("gfx/general/ob_mainmenu_background.png");
	thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgb,
				boost::ref(m_texture_menu), boost::ref(img_menu), false));

	gfx::ImageRGBASptr img_particle[OB_PARTICLE_COUNT];
	const char *fname_particle[OB_PARTICLE_COUNT] =
	{
		"gfx/billboards/particle_crackle_1.png",
		"gfx/billboards/particle_crackle_2.png",
		"gfx/billboards/particle_crackle_3.png",
		"gfx/billboards/particle_crackle_4.png",
		"gfx/billboards/particle_glow_medium.png",
		"gfx/billboards/particle_glow_sharp.png",
		"gfx/billboards/particle_glow_soft.png",
		"gfx/billboards/particle_shockwave_narrow_1.png",
		"gfx/billboards/particle_shockwave_narrow_2.png",
		"gfx/billboards/particle_shockwave_narrow_3.png",
		"gfx/billboards/particle_shockwave_narrow_4.png",
		"gfx/billboards/particle_shockwave_wide_1.png",
		"gfx/billboards/particle_shockwave_wide_2.png",
		"gfx/billboards/particle_shockwave_wide_3.png",
		"gfx/billboards/particle_shockwave_wide_4.png",
		"gfx/billboards/particle_smoke_hard_1.png",
		"gfx/billboards/particle_smoke_hard_2.png",
		"gfx/billboards/particle_smoke_hard_3.png",
		"gfx/billboards/particle_smoke_hard_4.png",
		"gfx/billboards/particle_smoke_soft_1.png",
		"gfx/billboards/particle_smoke_soft_2.png",
		"gfx/billboards/particle_smoke_soft_3.png",
		"gfx/billboards/particle_smoke_soft_4.png",
		"gfx/billboards/particle_sparkle_1.png",
		"gfx/billboards/particle_sparkle_2.png",
		"gfx/billboards/particle_sparkle_3.png",
		"gfx/billboards/particle_sparkle_4.png",
		"gfx/billboards/particle_sparkle_5.png",
		"gfx/billboards/particle_sparkle_6.png"
	};
	for(unsigned ii = 0; (ii < OB_PARTICLE_COUNT); ++ii)
	{
		img_particle[ii] = gfx::ImageRGBASptr(new gfx::ImageRGBA(fname_particle[ii]));
		thr::dispatch_privileged(boost::bind(&Globals::reserve_texture_rgba,
					boost::ref(m_texture_particle[ii]), boost::ref(*(img_particle[ii])), true));
	}

	m_sample_alarm = snd::SampleSptr(new snd::Sample("snd/ob_alarm.wav", 0.5f));
	m_sample_alarm_over = snd::SampleSptr(new snd::Sample("snd/ob_alarm_over.wav", 0.5f));
	m_sample_contact = snd::SampleSptr(new snd::Sample("snd/ob_contact.wav", 0.5f));
	m_sample_flak = snd::SampleSptr(new snd::Sample("snd/ob_flak.wav", 0.5f));
	m_sample_locked = snd::SampleSptr(new snd::Sample("snd/ob_locked.wav", 0.5f));
	m_sample_nuke = snd::SampleSptr(new snd::Sample("snd/ob_nuke.wav", 0.5f));
	m_sample_railgun = snd::SampleSptr(new snd::Sample("snd/ob_railgun.wav", 0.5f));
	m_sample_route_change = snd::SampleSptr(new snd::Sample("snd/ob_route_change.wav", 0.5f));
	m_sample_route_change_accepted = snd::SampleSptr(new snd::Sample("snd/ob_route_change_accepted.wav", 0.5f));
	m_sample_target_destroyed = snd::SampleSptr(new snd::Sample("snd/ob_target_destroyed.wav", 0.5f));

	m_mesh_atmosphere = gfx::MeshSptr(new Atmosphere(subdivide - 2, coalesce));
	m_mesh_bullet_flak = gfx::MeshSptr(new VisualizationFlak());
	m_mesh_bullet_railgun = gfx::MeshSptr(new VisualizationRailgun());
	m_mesh_missile_anti = gfx::MeshSptr(new gfx::MeshStatic("mdl/missile_anti.xml"));
	m_mesh_missile_nuke = gfx::MeshSptr(new gfx::MeshStatic("mdl/missile_icbm.xml"));
	m_mesh_nuke_marker = gfx::MeshSptr(new VisualizationNuke());
	m_mesh_orbit = gfx::MeshSptr(new VisualizationOrbit());
	m_mesh_silo = gfx::MeshSptr(new gfx::MeshStatic("mdl/silo.xml"));
	m_mesh_skybox_0 = gfx::MeshSptr(new VisualizationMesh("mdl/skybox_back.xml"));
	m_mesh_skybox_1 = gfx::MeshSptr(new VisualizationMesh("mdl/skybox_bottom.xml"));
	m_mesh_skybox_2 = gfx::MeshSptr(new VisualizationMesh("mdl/skybox_front.xml"));
	m_mesh_skybox_3 = gfx::MeshSptr(new VisualizationMesh("mdl/skybox_left.xml"));
	m_mesh_skybox_4 = gfx::MeshSptr(new VisualizationMesh("mdl/skybox_right.xml"));
	m_mesh_skybox_5 = gfx::MeshSptr(new VisualizationMesh("mdl/skybox_top.xml"));
	m_mesh_planet = gfx::MeshSptr(new Planet(subdivide, coalesce, texsize, volsize,
				&m_height_map_planet, generate_enabled));

	const char *fname_siege_cruiser[] =
	{
		"mdl/siegecruiser_0.xml",
		"mdl/siegecruiser_1.xml",
		"mdl/siegecruiser_2.xml",
		"mdl/siegecruiser_3.xml",
		"mdl/siegecruiser_4.xml",
		"mdl/siegecruiser_5.xml",
		"mdl/siegecruiser_6.xml"
	};
	for(unsigned ii = 0; (ii < OB_SIEGE_CRUISER_MESH_COUNT); ++ii)
	{
		gfx::Mesh *sc = new gfx::MeshStatic();
		sc->load(fname_siege_cruiser[ii], 0.016f);
		m_mesh_siege_cruiser[ii] = gfx::MeshSptr(sc);
	}

	_precalculated = true;
}

void Globals::reserve_shader(gfx::ShaderSptr &dst, const char *cstr)
{
	dst = gfx::ShaderSptr(new gfx::Shader(cstr));
}

void Globals::reserve_texture_rgb(gfx::Texture2DSptr &dst,
		const gfx::ImageRGB &img, bool clamp)
{
	dst = gfx::Texture2DSptr(new gfx::Texture2D(img, clamp));
}

void Globals::reserve_texture_rgba(gfx::Texture2DSptr &dst,
		const gfx::ImageRGBA &img, bool clamp)
{
	dst = gfx::Texture2DSptr(new gfx::Texture2D(img, clamp));
}

void Globals::unreserve()
{
	_precalculated = false;
	m_shader_object = gfx::ShaderSptr();
	m_shader_overlay = gfx::ShaderSptr();
	m_shader_overlay_billboard = gfx::ShaderSptr();
	m_shader_overlay_line = gfx::ShaderSptr();
	m_shader_planet = gfx::ShaderSptr();
	m_mesh_atmosphere = gfx::MeshSptr();
	m_mesh_bullet_flak = gfx::MeshSptr();
	m_mesh_bullet_railgun = gfx::MeshSptr();
	m_mesh_missile_anti = gfx::MeshSptr();
	m_mesh_missile_nuke = gfx::MeshSptr();
	m_mesh_nuke_marker = gfx::MeshSptr();
	m_mesh_orbit = gfx::MeshSptr();
	m_mesh_planet = gfx::MeshSptr();
	m_mesh_skybox_0 = gfx::MeshSptr();
	m_mesh_skybox_1 = gfx::MeshSptr();
	m_mesh_skybox_2 = gfx::MeshSptr();
	m_mesh_skybox_3 = gfx::MeshSptr();
	m_mesh_skybox_4 = gfx::MeshSptr();
	m_mesh_skybox_5 = gfx::MeshSptr();
	m_mesh_silo = gfx::MeshSptr();
	BOOST_FOREACH(gfx::MeshSptr &vv, m_mesh_siege_cruiser)
	{
		vv = gfx::MeshSptr();
	}
	m_texture_flak_ammo = gfx::Texture2DSptr();
	m_texture_icon_bullet_flak = gfx::Texture2DSptr();
	m_texture_icon_bullet_railgun = gfx::Texture2DSptr();
	m_texture_icon_city = gfx::Texture2DSptr();
	m_texture_icon_missile_anti_nuke = gfx::Texture2DSptr();
	m_texture_icon_missile_anti_ship = gfx::Texture2DSptr();
	m_texture_icon_silo = gfx::Texture2DSptr();
	m_texture_missile_anti_nuke = gfx::Texture2DSptr();
	m_texture_missile_anti_ship = gfx::Texture2DSptr();
	m_texture_missile_nuke = gfx::Texture2DSptr();
	m_texture_railgun_ammo = gfx::Texture2DSptr();
	m_texture_silo_anti_nuke = gfx::Texture2DSptr();
	m_texture_silo_anti_ship = gfx::Texture2DSptr();
	m_texture_silo_both = gfx::Texture2DSptr();
	m_texture_target = gfx::Texture2DSptr();
	m_texture_trace = gfx::Texture2DSptr();
	BOOST_FOREACH(gfx::Texture2DSptr &vv, m_texture_particle)
	{
		vv = gfx::Texture2DSptr();
	}
}

SDL_Cursor* Globals::create_cursor_blank()
{
	uint8_t cdata[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
	return SDL_CreateCursor(cdata, cdata, 8, 1, 0, 0);
}

void ob::glob_init(const gfx::SurfaceScreen &pscreen,
		const char *pdetail)
{
	glob_quit();
	ui::log.connect(ui::log_default);
	glob = new Globals(pscreen, pdetail);
	ui::log.connect(log_console);
}

void ob::glob_init(const gfx::SurfaceScreen &pscreen,
		const std::string &pdetail)
{
	glob_init(pscreen, pdetail.c_str());
}

void ob::glob_precalc()
{
	if(!glob)
	{
		std::stringstream sstr;
		sstr << "globals not initialized";
		BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
	}
	glob->precalc();
}

void ob::glob_quit()
{
	delete glob;
	glob = NULL;
	ui::log.disconnect_all_slots();
}

void ob::glob_set_game(Game *op)
{
	if(op)
	{
		if(game)
		{
			std::stringstream sstr;
			sstr << "trying to set game when previous exists";
			BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
		}
		game = op;
	}
	else
	{
		game = NULL;
	}
}

