#include "ob_menu.hpp"

#include "math/generic.hpp"
#include "snd/generic.hpp"
#include "ui/generic.hpp"
#include "ob_constants.hpp"
#include "ob_settings.hpp"

using namespace ob;

static const char *resolutions[] =
{
	"640x480@32",
	"800x600@32",
	"1024x768@32",
	"1280x720@32",
	"1280x960@32",
	"1280x1024@32",
	"1366x768@32",
	"1600x1200@32",
	"1680x1050@32",
	"1920x1080@32",
	"1920x1200@32",
	NULL
};

/** Maximum focus time. */
static const int FOCUS_TIME_MAX = 40;

/** Maximum focus time as float. */
static const float FOCUS_TIME_MAX_FLOAT = static_cast<float>(FOCUS_TIME_MAX);

Menu::Menu(const char *name, MenuEnum func, Menu *parent) :
	m_parent(parent),
	m_name_utf8(name),
	m_name_wide(ui::wstr_utf8(name)),
	m_focus_time(0),
	m_focus_time_float(0.0f),
	m_func(func)
{
	this->updateText(0);
}

void Menu::add(Menu *op)
{
	m_recursive.push_back(MenuSptr(op));
}

float Menu::decFocusTime()
{
	m_focus_time = math::max(m_focus_time - 1, 0);
	m_focus_time_float = static_cast<float>(m_focus_time) / FOCUS_TIME_MAX_FLOAT;
	return m_focus_time_float;
}

float Menu::incFocusTime()
{
	m_focus_time = math::min(m_focus_time + 1, FOCUS_TIME_MAX);
	m_focus_time_float = static_cast<float>(m_focus_time) / FOCUS_TIME_MAX_FLOAT;
	return m_focus_time_float;
}

gfx::Color Menu::getFocusColor(float alpha)
{
	gfx::Color ret = OB_COLOR_ACTIVE;
	float mul = m_focus_time_float * 0.5f + 0.5f;
	ret.r() *= mul;
	ret.g() *= mul;
	ret.b() *= mul;
	ret.a() *= alpha;
	return ret;
}

void Menu::updateText(int op)
{
	switch(m_func)
	{
		case FULLSCREEN:
			{
				settingi &fs = conf->getFullscreen();
				if(op != 0)
				{
					fs.set(op);
				}
				m_name_utf8 = (fs.get() > 0) ?
					std::string("Fullscreen") :
					std::string("Windowed");
				m_name_wide = ui::wstr_utf8(m_name_utf8);
			}
			break;

		case INVERT_MOUSE:
			{
				settingf &inv = conf->getCameraRotSpeedX();
				if(op != 0)
				{
					inv.set(-op * conf->getCameraRotSpeedY().get());
				}
				m_name_utf8 = (inv.get() >= 0.0f) ?
					std::string("Invert mouse off") :
					std::string("Invert mouse on");
				m_name_wide = ui::wstr_utf8(m_name_utf8);
			}
			break;

		case RESOLUTION:
			{
				std::string res = conf->getResolution();
				int ii = 0;
				for(; true; ++ii)
				{
					if(!resolutions[ii])
					{
						ii = 0;
						break;
					}
					if(res.compare(resolutions[ii]) == 0)
					{
						break;
					}
				}
				ii += op;
				if(ii < 0)
				{
					ii = 0;
				}
				if(!resolutions[ii])
				{
					--ii;
				}
				m_name_utf8.assign(resolutions[ii]);
				m_name_wide = ui::wstr_utf8(m_name_utf8);
				conf->setResolution(m_name_utf8);
			}
			break;

		case VOLUME_MUSIC:
			{
				settingf &vol = conf->getVolumeSamples();
				vol.set(vol.get() + 0.1f * op);
				std::stringstream sstr;
				sstr << "music volume: " << math::lround(vol.get() * 100);
				m_name_utf8.assign(sstr.str());
				m_name_wide = ui::wstr_utf8(m_name_utf8);
				snd::volume_music = vol.get();
			}
			break;

		case VOLUME_SAMPLES:
			{
				settingf &vol = conf->getVolumeSamples();
				vol.set(vol.get() + 0.1f * op);
				std::stringstream sstr;
				sstr << "sample volume: " << math::lround(vol.get() * 100);
				m_name_utf8.assign(sstr.str());
				m_name_wide = ui::wstr_utf8(m_name_utf8);
				snd::volume_samples = vol.get();
			}
			break;

		default:
			break;
	}
}

