#include "ob_menu_state.hpp"

#include "snd/generic.hpp"
#include "thr/dispatch.hpp"
#include "ui/ui_stack.hpp"
#include "ob_constants.hpp"
#include "ob_game.hpp"
#include "ob_globals.hpp"
#include "ob_high_score_state.hpp"
#include "ob_menu.hpp"

using namespace ob;

/** Focused menu item. */
static const float FONT_SIZE_MAX = 1.25 * OB_FONT_SIZE;

/** Focused menu item. */
static const float FONT_SIZE_TITLE = OB_FONT_SIZE * 1.67;

/** Gap value. */
static const float TITLE_BOTTOM = FONT_SIZE_TITLE * 0.2f;

/** Help filenames. */
static const char *help_filenames[] =
{
	"gfx/general/instructions_1.png",
	"gfx/general/instructions_2.png",
	"gfx/general/instructions_3.png",
	"gfx/general/instructions_4.png",
	"gfx/general/instructions_5.png"
};

/** \brief Collect text lengths.
 *
 * @param tgt Target vector.
 * @param src Source menus.
 * @param global Global multiplier.
 */
void collect_widths(std::vector<float> &tgt, const std::vector<MenuSptr> &src,
		float global)
{
	BOOST_FOREACH(const MenuSptr &vv, src)
	{
		const Menu *cmenu = vv.get();
		float fsize = math::mix(OB_FONT_SIZE, FONT_SIZE_MAX, cmenu->getFocusTimeFloat());
		int idx;
		float width;
		boost::tie(idx, width) = glob->getFont().calcLineWidth(fsize, cmenu->getText());
		tgt.push_back(width * global);
	}
}

/** \brief Draw a menu.
 *
 * @param wid Width vector.
 * @param mvec Menu vector.
 * @param global Global multiplier.
 * @param maxwid Maximum width encountered.
 */
void draw_menu_list(std::vector<float> &wid, const std::vector<MenuSptr> &mvec,
		float global, float maxwid, const math::rect2f &sarea)
{
	float FONT_TOP = mvec.size() * (FONT_SIZE_MAX + OB_FONT_GAP) + OB_FONT_GAP;

	for(unsigned ii = 0; (ii < mvec.size()); ++ii)
	{
		gfx::Color mcolor = mvec[ii]->getFocusColor(global);
		gfx::draw_fill(0, mcolor);
		gfx::draw_fill(1, mcolor);
		gfx::draw_fill(2, mcolor);
		gfx::draw_fill(3, mcolor);

		gfx::draw_text_line(sarea.w() - OB_FONT_GAP - maxwid * global + (maxwid - wid[ii]),
				FONT_TOP - (ii + 1) * (FONT_SIZE_MAX + OB_FONT_GAP) + OB_FONT_BOTTOM,
				math::mix(OB_FONT_SIZE, FONT_SIZE_MAX, mvec[ii]->getFocusTimeFloat()),
				mvec[ii]->getText(), glob->getFont());
	}
}

MenuState::MenuState(Menu *menu, unsigned idx) :
	m_menu(menu),
	m_menu_curr(menu),
	m_menu_prev(NULL),
	m_index(idx),
	m_help_page(-1)
{
	m_title.set(m_menu->getText());
	snd::play_stream("snd/music_menu.ogg");
}

MenuState::~MenuState()
{
	thr::wait_privileged(boost::bind(&MenuState::taskTexture, this,
				reinterpret_cast<char*>(NULL)));
}

void MenuState::activate(Menu *op, int mode, ui::UiStack &st)
{
	if((0 == mode) && (op->getNumRecursive() > 0))
	{
		this->descend(op);
	}

	if(op->getFunc() == HELP)
	{
		if(0 > m_help_page)
		{
			m_help_page = 0;
		}
		else
		{
			m_help_page = math::max(math::min(m_help_page + mode, static_cast<int>(OB_HELP_PAGE_COUNT - 1)), 0);
		}
		thr::wait_privileged(boost::bind(&MenuState::taskTexture, this,
					help_filenames[m_help_page]));
		return;
	}
	m_help_page = -1;
	thr::wait_privileged(boost::bind(&MenuState::taskTexture, this,
				reinterpret_cast<char*>(NULL)));

	switch(op->getFunc())
	{
		case HIGH_SCORES:
			if(0 == mode)
			{
				st.pushState(new HighScoreState());
			}
			break;

		case QUIT:
			if(0 == mode)
			{
				_alive = false;
			}
			break;

		case START:
			if(0 == mode)
			{
				st.pushState(new Game());
				snd::play_stream("snd/music_game.ogg");
			}
			break;

		default:
			op->updateText(mode);
			break;
	}
}

void MenuState::ascend()
{
	this->descend(m_menu_curr->getParent());
}

void MenuState::checkIndex()
{
	if(m_index >= static_cast<int>(m_menu_curr->getNumRecursive()))
	{
		--m_index;
	}
	if(0 > m_index)
	{
		m_index = 0;
	}
}

void MenuState::descend(Menu *op)
{
	m_menu_prev = m_menu_curr;
	m_menu_curr = op;
	// Sanity check.
	if(NULL == m_menu_curr)
	{
		m_menu_curr = m_menu;
	}
	// Check if no move.
	if(m_menu_prev == m_menu_curr)
	{
		m_menu_prev = NULL;
	}
	else
	{	
		m_menu_curr->setFocusTime(0);
		m_title.set(m_menu_curr->getText());
	}
	this->checkIndex();
}

bool MenuState::handle(ui::EventKey &ev, ui::UiStack &st)
{
	if(ev.isPress())
	{
		switch(ev.getCode())
		{
			case SDLK_RETURN:
				this->activate(m_menu_curr->getChild(m_index), 0, st);
				break;

			case SDLK_BACKSPACE:
			case SDLK_WORLD_7:
				this->ascend();
				break;

			case SDLK_ESCAPE:
				if(NULL == m_menu_curr->getParent())
				{
					_alive = false;
				}
				else
				{
					this->ascend();
				}
				break;

			case SDLK_LEFT:
				this->activate(m_menu_curr->getChild(m_index), -1, st);
				break;

			case SDLK_RIGHT:
				this->activate(m_menu_curr->getChild(m_index), 1, st);
				break;

			case SDLK_UP:
				--m_index;
				break;
				
			case SDLK_DOWN:
				++m_index;
				this->checkIndex();
				break;

			case SDLK_HOME:
				m_index = 0;
				break;

			case SDLK_END:
				m_index = m_menu->getNumRecursive();
				if(0 < m_index)
				{
					--m_index;
				}
				break;

			default:
				break;
		}
	}

	return true;
}

bool MenuState::handle(ui::EventMisc &ev, ui::UiStack &st)
{
	return true;
}

bool MenuState::handle(ui::EventMouseButton &ev, ui::UiStack &st)
{
	return true;
}

bool MenuState::handle(ui::EventMouseMotion &ev, ui::UiStack &st)
{
	return true;
}

void MenuState::reinstate(ui::UiStack &st)
{
	ui::key_repeat_on();
}

void MenuState::taskRender(gfx::SurfaceScreen &screen) const
{
	math::rect2f sarea = screen.getArea();

	screen.select2D();
	screen.clear(true, true);

	gfx::bind_shader_2d_texture();
	gfx::draw_rect_textured_fill(sarea, gfx::Color(1.0f, 1.0f, 1.0f, 1.0f),
			glob->getTextureMenu());

	std::vector<float> currlen, prevlen;
	float maxwidth_prev = FLT_MIN;
	if(m_menu_prev)
	{
		collect_widths(prevlen, m_menu_prev->getRecursive(), m_menu_prev->getFocusTimeFloat());
		BOOST_FOREACH(float vv, prevlen)
		{
			maxwidth_prev = math::max(maxwidth_prev, vv);
		}
	}
	float maxwidth_curr = FLT_MIN;
	collect_widths(currlen, m_menu_curr->getRecursive(), m_menu_curr->getFocusTimeFloat());
	BOOST_FOREACH(float vv, currlen)
	{
		maxwidth_curr = math::max(maxwidth_curr, vv);
	}
	float maxwidth = math::max(maxwidth_curr, maxwidth_prev);

	screen.select2D();
	gfx::bind_shader_2d();
	math::vec2f corner(sarea.w() - maxwidth - OB_FONT_GAP * 2.0f,
			sarea.h() - FONT_SIZE_TITLE - OB_FONT_GAP * 2.0f);
	gfx::draw_rect(0.0f, corner.y(), sarea.w(), sarea.h() - corner.y(), OB_COLOR_SHADE);
	gfx::draw_rect(corner.x(), 0.0f, sarea.w() - corner.x(), corner.y());
	gfx::draw_fill(0, 0.0f, corner.y());
	gfx::draw_fill(1, corner.x(), corner.y());
	gfx::draw_fill(2, corner.x(), 0.0f);
	gfx::draw_fill(0, OB_COLOR_MENU_BORDER);
	gfx::draw_fill(1, OB_COLOR_MENU_BORDER);
	gfx::draw_fill(2, OB_COLOR_MENU_BORDER);
	glLineWidth(OB_LINE_WIDTH);
	gfx::draw_arrays(GL_LINE_STRIP, 3);
	glLineWidth(1.0f);
	//gfx::draw_line(corner.x(), corner.y(), 0.0f, corner.y(), OB_COLOR_BORDER);
	//gfx::draw_line(corner.x(), corner.y(), corner.x(), 0.0f);

	gfx::bind_shader_2d_font();
	if(m_menu_prev)
	{
		draw_menu_list(prevlen, m_menu_prev->getRecursive(),
				m_menu_prev->getFocusTimeFloat(), maxwidth_prev, sarea);
	}

	draw_menu_list(currlen, m_menu_curr->getRecursive(),
			m_menu_curr->getFocusTimeFloat(), maxwidth_curr, sarea);

	gfx::draw_fill(0, OB_COLOR_TITLE);
	gfx::draw_fill(1, OB_COLOR_TITLE);
	gfx::draw_fill(2, OB_COLOR_TITLE);
	gfx::draw_fill(3, OB_COLOR_TITLE);
	gfx::draw_text_line(OB_FONT_GAP, sarea.h() - FONT_SIZE_TITLE - OB_FONT_GAP + TITLE_BOTTOM,
			FONT_SIZE_TITLE, m_title.get(), glob->getFont());

	if(NULL != m_overlay.get())
	{
		gfx::bind_shader_2d_texture();
		math::rect2f harea(0.0f, 0.0f, corner.x(), corner.y());
		gfx::draw_rect_textured_fit(harea, gfx::Color(1.0f, 1.0f, 1.0f, 1.0f),
				*m_overlay);
	}

	screen.update();
}

void MenuState::taskTexture(const char *pfname)
{
	if(NULL != pfname)
	{
		m_overlay = gfx::Texture2DSptr(new gfx::Texture2D(pfname));
	}
	else
	{
		m_overlay = gfx::Texture2DSptr();
	}
}

void MenuState::update(gfx::SurfaceScreen &screen, ui::UiStack &st,
		bool prender)
{
	m_menu_curr->incFocusTime();
	if(NULL != m_menu_prev)
	{
		float focus_prev = m_menu_prev->decFocusTime();
		if(focus_prev <= 0.0f)
		{
			m_menu_prev = NULL;
		}
	}

	std::vector<MenuSptr> &mvec = m_menu_curr->getRecursive();
	for(unsigned ii = 0; (ii < mvec.size()); ++ii)
	{
		Menu *cmenu = mvec[ii].get();
		// Hack: do not increment or modify previous focus menu ever.
		if(cmenu != m_menu_prev)
		{
			if(static_cast<int>(ii) == m_index)
			{
				cmenu->incFocusTime();
			}
			else
			{
				cmenu->decFocusTime();
			}
		}
	}

	m_title.update();

	if(prender)
	{
		thr::wait_privileged(boost::bind(&MenuState::taskRender, this,
					boost::ref(screen)));
	}
}

Menu* MenuState::create_menu_game()
{
	return NULL;
}

Menu* MenuState::create_menu_main()
{
	Menu *ret = new Menu("Orbital Bombardment"),
			 *options = new Menu("Options", NONE, ret);

	//options->add(new Menu("Detail level", NONE, options));
	options->add(new Menu("Invert mouse", INVERT_MOUSE, options));
	options->add(new Menu("Resolution", RESOLUTION, options));
	options->add(new Menu("Music volume", VOLUME_MUSIC, options));
	options->add(new Menu("Samples volume", VOLUME_SAMPLES, options));
	//options->add(new Menu("Sensitivity", NONE, options));

	ret->add(new Menu("Start", START, ret));
	ret->add(options);
	ret->add(new Menu("instructions", HELP, ret));
	ret->add(new Menu("Records", HIGH_SCORES, ret));
	ret->add(new Menu("Exit", QUIT, ret));
	return ret;
}

