#include "ob_missile_anti.hpp"

#include "math/random.hpp"
#include "ob_constants.hpp"
#include "ob_game.hpp"
#include "ob_globals.hpp"

using namespace ob;

/** Missile collision area. */
static const float OB_COLLISION_MISSILE_ANTI = 1.0f;

/** Missile main acceleration. */
static const float OB_MISSILE_ACCELERATION = 10.0f;

/** Missile lateral acceleration. */
static const float OB_MISSILE_LATERAL_ACCELERATION = 0.25f;

/** Anti-nuke alarm proximity. */
static const float OB_MISSILE_ANTI_NUKE_ALARM_PROXIMITY = 600.0f;

/** Convenience. */
static const float OB_MISSILE_ANTI_NUKE_ALARM_PROXIMITY_2 = OB_MISSILE_ANTI_NUKE_ALARM_PROXIMITY * OB_MISSILE_ANTI_NUKE_ALARM_PROXIMITY;

/** Missile alarm when closer than this. */
static const float OB_MISSILE_ANTI_SHIP_ALARM_PROXIMITY = 800.0f;

/** Convenience. */
static const float OB_MISSILE_ANTI_SHIP_ALARM_PROXIMITY_2 = OB_MISSILE_ANTI_SHIP_ALARM_PROXIMITY * OB_MISSILE_ANTI_SHIP_ALARM_PROXIMITY;

/** Warning blinking level. */
static const int OB_WARNING_BLINK_LEVEL = 3;

/** Missile speed. */
static const float OB_MISSILE_SPEED = 90.0f;

MissileAnti::MissileAnti(const math::vec3d &pos, const math::vec3d &dir,
		const Missile *tgt) :
	Missile(pos, dir, glob->getMeshMissileAnti(), OB_COLLISION_MISSILE_ANTI,
			OB_FACTION_ENEMY_MISSILE,
			((NULL != tgt) ? &(glob->getTextureMissileAntiNuke()) : &(glob->getTextureMissileAntiShip()))),
	m_target(tgt) { }

math::vec3d MissileAnti::getTargetPos() const
{
	if(m_target)
	{
		if(game->hasMissileNuke(m_target))
		{
			return m_target->getPos();
		}
		else
		{
			return m_rot;
		}
	}
	return game->getView().getPos();
}

bool MissileAnti::update()
{
	if(this->isDead())
	{
		return this->updateDead();
	}

	math::vec3d tgt_pos;
	math::vec3d tgt_dir;
	if(m_target)
	{
		if(game->hasMissileNuke(m_target))
		{
			tgt_pos = m_target->getPos();
			tgt_dir = m_target->getRot();

			if(math::length2(tgt_pos - m_pos) <= OB_MISSILE_ANTI_NUKE_ALARM_PROXIMITY_2)
			{
				this->setBillboard(&(glob->getTextureMissileAntiNukeWarning()));
				this->enableRandom(OB_WARNING_BLINK_LEVEL);
			}
		}
		else
		{
			this->gamisticEffect(NULL);
			return this->updateDead();
		}
	}
	else
	{
		const GameView &gv = game->getView();
		tgt_pos = gv.getPos();
		tgt_dir = gv.getDir() * OB_ORBIT_SPEED;

		// Play alarm if dangerously close.
		if(math::length2(tgt_pos - m_pos) <= OB_MISSILE_ANTI_SHIP_ALARM_PROXIMITY_2)
		{
			if(!m_alarm_sounded)
			{
				this->setBillboard(&(glob->getTextureMissileAntiShipWarning()));
				snd::play(glob->getSampleAlarm(), math::vec3f(m_pos));
				m_alarm_sounded = true;
			}
			this->enableRandom(OB_WARNING_BLINK_LEVEL);
		}
	}

	math::vec3d dir = calculateMissileAimDirection(tgt_pos, tgt_dir, m_pos, m_rot);
	math::vec3d udir = updatePosDir(m_pos, m_rot, dir, OB_MISSILE_SPEED,
			OB_MISSILE_ACCELERATION, OB_MISSILE_LATERAL_ACCELERATION);
	m_wm.loadLookAt(math::vec3f(m_pos), math::vec3f(m_pos + udir), math::vec3f(dir));

	this->updateCollisionData(math::vec3f(m_pos));
	this->updateAreas(game->getOctree());
	CollisionElement *other = this->checkCollisions();
	if(other)
	{
		//std::cout << "collision!\n";
		other->gamisticEffect(NULL);
		this->gamisticEffect(NULL);
		return true;
	}

	// 50% chance of either particle trail.
	if(m_age%2==0)
	{
		if(!math::mrand(0, 2))
		{
			Particle *par = new Particle(gfx::Color(0.5f, 0.55f, 0.55f, 0.4f),
					math::vec3f(m_pos)-math::normalize(m_rot)*0.7, OB_BILLBOARD_SIZE_SMOKE,
					math::vec3f(0.0f, 0.0f, 0.0f), OB_PARTICLE_TIME_SMOKE, -OB_BILLBOARD_SIZE_SMOKE/2.0f);
			game->addParticle(SMOKE_SOFT_1, par);
		}
		else
		{
			Particle *par = new Particle(gfx::Color(1.0f, 1.0f, 0.5f, 0.5f),
					math::vec3f(m_pos)-math::normalize(m_rot)*0.7, OB_BILLBOARD_SIZE_SMOKE,
					math::vec3f(0.0f, 0.0f, 0.0f), OB_PARTICLE_TIME_SMOKE, -OB_BILLBOARD_SIZE_SMOKE);
			game->addParticle(SMOKE_HARD_1, par);
		}
	}

	this->updateVisibility(true);
	if(!this->incrementAge(OB_BULLET_DEATH_PROBABILITY))
	{
		this->gamisticEffect(NULL);
	}
	return true;
}

