#ifndef OB_TARGET_HPP
#define OB_TARGET_HPP

#include "math/vec.hpp"

namespace gfx
{
	class EntityObject;
	class Texture2D;
}

namespace ob
{
	/** 2D priority (target).
	 *
	 * Used for selecting targets in the game in 2D space.
	 */
	class Target
	{
		private:
			/** Target position in 2D. */
			math::vec2f m_pos;

			/** Reference to target. */
			const gfx::EntityObject *m_target;

			/** Billboard size. */
			float m_billboard_size;

			/** Corresponding target icon. */
			const gfx::Texture2D *m_icon;

		public:
			/** \brief Accessor.
			 *
			 * @return Billboard size.
			 */
			inline float getBillboardSize() const
			{
				return m_billboard_size;
			}

			/** \brief Accessor.
			 *
			 * @return Texture pointer.
			 */
			inline const gfx::Texture2D* getIcon() const
			{
				return m_icon;
			}

			/** \brief Accessor.
			 *
			 * @return Position.
			 */
			inline const math::vec2f& getPos() const
			{
				return m_pos;
			}

			/** \brief Accessor.
			 *
			 * @return Target.
			 */
			inline const gfx::EntityObject* getTarget() const
			{
				return m_target;
			}

		public:
			/** \brief Constructor.
			 *
			 * @param pos Position (in 2D).
			 * @param target Target.
			 */
			Target(const math::vec2f &pos, const gfx::EntityObject *target,
					float billboard_size, const gfx::Texture2D &icon) :
				m_pos(pos),
				m_target(target),
				m_billboard_size(billboard_size),
				m_icon(&icon)	{ }

			/** \brief Destructor. */
			~Target() { }

		public:
			/** \brief Less than.
			 *
			 * @param rhs Right-hand-side operand.
			 */
			inline bool operator<(const Target &rhs) const
			{
				return (math::length2(m_pos) > math::length2(rhs.m_pos));
			}
	};

	/** Convenience typedef. */
	typedef boost::shared_ptr<Target> TargetSptr;
}

#endif
