#include "snd/sample.hpp"

#include "data/generic.hpp"
#include "ui/generic.hpp"

using namespace snd;

Sample::Sample() :
	m_buffer(0) { }

Sample::Sample(const char *pfname, float level) :
	m_buffer(0),
	m_level(level)
{
	this->load(std::string(pfname));
}

Sample::Sample(const std::string &pfname, float level) :
	m_buffer(0),
	m_level(level)
{
	this->load(pfname);
}

Sample::Sample(const uint8_t *data, unsigned size, unsigned freq,
		float level) :
	m_buffer(0),
	m_level(level)
{
	this->generateBuffer();
	alBufferData(m_buffer, AL_FORMAT_STEREO16, data, size, freq);
}

Sample::~Sample()
{
	this->unreserve();
}

void Sample::generateBuffer()
{
	alGenBuffers(1, &m_buffer);
	if(0 == m_buffer)
	{
		std::stringstream sstr;
		sstr << "could not create audio buffer";
		BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
	}
}

void Sample::load(const std::string &pfname)
{
	this->unreserve();

	//ui::log(std::string("loading sample '") + pfname + std::string("'"));
	std::string filename = data::open_search(pfname);

	this->generateBuffer();

	SDL_AudioSpec spec;
	uint8_t *buf;
	uint32_t len;

	if(&spec != SDL_LoadWAV(filename.c_str(), &spec, &buf, &len))
	{
		std::stringstream sstr;
		sstr << "could not load wav file: '" << filename << "': " <<
			SDL_GetError();
		BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
	}
	ALenum format;
	switch(spec.format)
	{
		case AUDIO_U8:
			format = (spec.channels > 1) ? AL_FORMAT_STEREO8 : AL_FORMAT_MONO8;
			break;

		case AUDIO_S8:
			format = (spec.channels > 1) ? AL_FORMAT_STEREO8 : AL_FORMAT_MONO8;
			break;

		case AUDIO_U16SYS:
			format = (spec.channels > 1) ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16;
			break;
	
		case AUDIO_S16SYS:
			format = (spec.channels > 1) ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16;
			break;

		default:
			{
				std::stringstream sstr;
				sstr << "unknown audio format: " << spec.format;
				BOOST_THROW_EXCEPTION(std::runtime_error(sstr.str()));
			}
			break;
	}

#if defined(DEBUG)
	std::cout << "Sample: " << size << " bytes, " << freq << " Hz, ";
	switch(format)
	{
		case AL_FORMAT_MONO8:
			std::cout << "8-bit mono";
			break;

		case AL_FORMAT_STEREO8:
			std::cout << "8-bit stereo";
			break;

		case AL_FORMAT_MONO16:
			std::cout << "16-bit mono";
			break;

		case AL_FORMAT_STEREO16:
			std::cout << "16-bit stereo";
			break;
	}
	std::cout << std::endl;
#endif

	alBufferData(m_buffer, format, buf, len, spec.freq);
	SDL_FreeWAV(buf);
}

void Sample::unreserve()
{
	if(0 < m_buffer)
	{
		alDeleteBuffers(1, &m_buffer);
		m_buffer = 0;
	}
}

