#ifndef SND_SAMPLE_HPP
#define SND_SAMPLE_HPP

#include "snd/generic.hpp"

#include <string>

namespace snd
{
	/** \brief One OpenAL sample.
	 */
	class Sample
	{
		private:
			/** Buffer for the wave. */
			ALuint m_buffer;

			/** Sound level. */
			float m_level;

		public:
			/** \brief Accessor.
			 *
			 * @return Buffer of this.
			 */
			inline ALuint getBuffer() const
			{
				return m_buffer;
			}

			/** \brief Accessor.
			 *
			 * @return Sound level.
			 */
			inline float getLevel() const
			{
				return m_level;
			}

			/** \brief Set the level.
			 *
			 * @param op New level.
			 */
			inline void setLevel(float op)
			{
				m_level = op;
			}

		public:
			/** \brief Empty constructor. */
			Sample();

			/** \brief Constructor.
			 *
			 * @param pfname Filename to load.
			 * @param level General level for this sample.
			 */
			Sample(const char *pfname, float level = 1.0f);

			/** \brief Constructor.
			 *
			 * @param pfname Filename to load.
			 * @param level General level for this sample.
			 */
			Sample(const std::string &pfname, float level = 1.0f);

			/** \brief Constructor.
			 *
			 * @param data Data count.
			 * @param size Size of data.
			 * @param freq Frequency.
			 * @param level General level for this sample.
			 */
			Sample(const uint8_t *data, unsigned size, unsigned freq, float level = 1.0f);

			/** \brief Destructor. */
			~Sample();

		private:
			/** \brief Generate the buffer in this sample.
			 */
			void generateBuffer();

		public:
			/** \brief Load a sample.
			 *
			 * Throws an exception on error.
			 *
			 * @param pfname Filename to load from.
			 */
			void load(const std::string &pfname);

		protected:
			/** \brief Release all resources reserved by this.
			 */
			void unreserve();
	};

	/** Convenience typedef. */
	typedef boost::shared_ptr<Sample> SampleSptr;
}

#endif
